  {----------------------------------------------------------------------}
  { Project title:       EMF52L                                          }
  { Project description: Software debugger for C51/C52 micro controllers }
  { Unit:                EMF52L                                          }
  { Unit Description:    Main program file                               }
  { Compiler:            Borland/Turbo Pascal 5.5, 6.0 or 7.0            }
  { Author:              Alexei A. Frounze (c) 1994...2003               }
  { Version and date:    2.30 april 2003                                 }
  { History:                                                             }
  {  2003-02-02 (feb 2)  minor changes, bugfixes in initialization code, }
  {                      compile-time support for different languages    }
  {                      added (English, Russian), made compilable with  }
  {                      Turbo Pascal 5.5, clock counter added           }
  {  2003-04-06 (apr 6)  added a workaround to the ms-windows problem:   }
  {                      windows could stop the debugger operation when  }
  {                      running a time-consuming piece of code, for     }
  {                      example a 5-seconds-long delay subroutine,      }
  {                      initiated by pressing P or G. Visually this     }
  {                      appeared as if the code took ifinite time to    }
  {                      complete and it could only be interrupted by    }
  {                      the space bar key. This happened because the    }
  {                      debugger asked BIOS if there's a keystroke ready}
  {                      too often. Windows decided that the debugger is }
  {                      an interactive application busy-wating on the   }
  {                      keyboard and paused execution of the debugger   }
  {                      until a key was pressed on the keyboard.        }
  {                      So, the code could be forced to complete by     }
  {                      pressing and releasing keys on the keyboard,    }
  {                      for example, the Ctrl key.                      }
  {----------------------------------------------------------------------}

{$M 16384,230000,230000}

Uses
  CRT, DOS,
  ProfText,
  S_Input,
  EmMCS52,
  EmF52L_U,
  EMF52L_L;

Const
  Max = 310000;
{  }
  CLineNormBack = Black;
  CLineNormText = Green;

  CLineCurrBack = Cyan;
  CLineCurrText = Black;

  CLineBreakBack = Red;
  CLineBreakText = Yellow;

  CLineCurrBreakBack = Red;
  CLineCurrBreakText = LightCyan;

{  }
  MemID     = 'EMF52L_2.0_MEMORY_FILE';

Type
  TableArr  = Array [0..ProgMax-1] of Word;

Var
  UpL, DnL,
  I,
  J     : Byte;
  C     : Char;
  F     : File;
  PCMem : Word;
  FList,
  PCF   : Boolean;
  ON,
  Name,
  MName : String;
  Table : ^TableArr;

Label
  Cycle,
  EmEnd;

Function FormatStr (S : String; Len : Byte) : String;
Begin
  While Length (S) < Len-1 do
    S := ' '+S+' ';
  If Length (S) < Len then S := S+' ';
  FormatStr := S
End;

Procedure OutError;
Var
  C : Char;
Begin
  Window (26, 10, 30, 6, MsgErrWinTitle, Red, White);
  GotoXY (30, 12);
  Write (MsgOutErr[Error]);
  GotoXY (30, 13);
  Write (FormatStr(ErrName, 22));
  While not KeyPressed do;
  While KeyPressed do C := ReadKey
End;

Procedure MakeTable;
Var
  F : File of Str78;
  S : Str78;
  N : Word;
Label
  LoopContinue1;
Begin
  FillChar (Table^, SizeOf(Table^), 0);
  ErrName := LN;
  Error := eRead;
  Assign (F, LN);
  {$I-} Reset (F); {$I+}
  If IOResult <> 0 then Exit;
  N := 0;
LoopContinue1:
  While not EOF (F) do
    Begin
      {$I-} Read (F, S); {$I+}
      If IOResult <> 0 then
        Begin
          Close (F);
          Exit
        End;
      Inc (N);
      If S[13] = ' ' then {Continue;} Goto LoopContinue1;
      Table^[Valu('$'+Copy(S,8,4))] := N
    End;
  Close (F);
  Error := eOk
End;

Procedure WriteWin;
Var I : Word;
Begin
  For I := 0 to WH-1 do
    Begin
      GotoXY (2, 1+I);
      Color (CLineNormBack, CLineNormText);
      If BFirst+WFirst+I = Table^[PC] then
        Color (CLineCurrBack, CLineCurrText);
      If NBreaks > 0 then
        Begin
          For J := 1 to NBreaks do
            If BFirst+WFirst+I = Table^[BP[J]] then
              Begin
                Color (CLineBreakBack, CLineBreakText);
                If BFirst+WFirst+I = Table^[PC] then
                  Color (CLineCurrBreakBack, CLineCurrBreakText)
              End
        End;
      If BFirst+WFirst+I <= StrNum then
        Write (FBuf^[WFirst+I]);
      ClrEol;
      Color (Black, Green);
      GotoXY (80, WhereY); Write (' ')
    End;
End;

Procedure X;
Var
  R,
  S : String;
  K : Word;
Label
  LoopBreak1;
Begin
  Color (Black, Green);
  GotoXY (1,1);
  PC2 := PC;
  For I := 1 to 14 do
    Begin
      If I = 1 then Color (CLineCurrBack, CLineCurrText)
      Else Color (CLineNormBack, CLineNormText);
      If NBreaks > 0 then
        For J := 1 to NBreaks do
          If BP [J] = PC2 then
            Begin
              If I = 1 then
                Color (CLineCurrBreakBack, CLineCurrBreakText)
              Else
                Color (CLineBreakBack, CLineBreakText);
              {Break} Goto LoopBreak1
            End;
LoopBreak1:
      K := PC2;
      S := Step(PC2);
      R := '   ';
      For K := K to PC2-1 do
        R := R+Hex(Prog^[K])+' ';
      While Length (R) < 15 do
        R := R + ' ';
      Insert (R, S, 6);
      GotoXY (2, WhereY);
      Write (S);
      ClrEol;
      GotoXY (80, WhereY);
      Color (Black, Green);
      Write (' ');
      If PC2 > ProgMax then PC2 := 0
    End;
  Color (Black, Green)
End;

Procedure XX;
Begin
  If FList then
    WriteWin
  Else
    X;
  Color (Black, Green)
End;

{$F+}
Procedure MakeScrXX; {Far;}
Begin
  MakeScr;
  XX;
End;
{$F-}

Procedure Help;
Var
  I : Byte;
  C : Char;
  Procedure WrSt (S : String);
  Var J : Byte;
  Begin
    Color (Cyan, Red);
    For J := 1 to Length (S) do
      Begin
        If S[J] = '-' then Color (Cyan, Black);
        Write (S[J]);
        If (S[J] = '') and (S[J+2] in [#33..#127]) then
          Color (Cyan, Red)
      End
  End;
Begin
  Window (3, 2, 76, 22, MsgHelpWinTitle, Cyan, Black);
  GotoXY (7, 4);
  Write (MsgHelpWinText[1]);
  For I := 2 to 18 do
    Begin
      GotoXY (7, 3+I);
      WrSt (MsgHelpWinText[I])
    End;
  While ReadKey <> #27 do;
  Delay (100);
  While KeyPressed do C := ReadKey;
  Color (Black, Green);
  ClrScr
End;

Procedure Breaks (PC : Word);
Var
  N : Byte;

  Procedure SetBreak (Ad : Word);
  Begin
    If NBreaks = 16 then Exit;
    Inc (NBreaks);
    BP [NBreaks] := Ad
  End;

  Procedure ClearBreak (Ad : Word);
  Var
    I : Byte;
  Label
    LoopBreak1;
  Begin
    If NBreaks = 0 then Exit;
    For I := 1 to NBreaks do
      If BP [I] = Ad then {Break;} Goto LoopBreak1;
  LoopBreak1:
    If I <> NBreaks then
      For I := I to NBreaks-1 do
        BP[I] := BP[I+1];
    Dec (NBreaks)
  End;

Begin
  If NBreaks > 0 then
    Begin
      For N := 1 to NBreaks do
        If BP [N] = PC then
          Begin
            ClearBreak (PC);
            Exit
          End;
      SetBreak (PC)
    End
  Else
    SetBreak (PC)
End;

Procedure Down;
Var S : String;
Begin
  If FList then
    Begin
      If BFirst+WFirst+WH-1 = StrNum then Exit;
      If WFirst+WH-1 < BLen-1 then
        Inc (WFirst)
      Else
        Begin
          NextBlock;
          If Error <> eOk then Exit
        End;
    End
  Else
    Begin
      PC2 := PC;
      S := Step (PC2);
      PC := PC2;
      If PC > ProgMax then PC := 0
    End
End;

Procedure Up;
Begin
  If FList then
    Begin
      If WFirst > 0 then
        Dec (WFirst)
      Else
        If BFirst <> 1 then
          Begin
            PrevBlock;
            If Error <> eOk then Exit;
          End
    End
  Else
    Begin
      Dec (PC);
      If PC > ProgMax then PC := ProgMax
    End
End;

Procedure SaveMemory;
Var
  S : String;
  F : File;
Begin
  Window (27, 11, 28, 5, MsgSaveMemWinTitle, LightGray, Black);
  S := MName;
  GotoXY (35, 13);
  CursorOn (UpL, DnL);
  ReadSK (S, 12);
  CursorOff;
  If S = '' then Exit;
  MName := S;
  S := MemID;
  ErrName := MName;
  Error := eWrite;
  Assign (F, MName);
  {$I-} Rewrite (F, 1); {$I+}
  If IOResult <> 0 then Exit;
  {$I-} BlockWrite (F, S, Length(MemID)+1); {$I+}
  If IOResult <> 0 then
    Begin
      Close (F);
      Exit
    End;
  {$I-} BlockWrite (F, PC, SizeOf(PC)); {$I+}
  If IOResult <> 0 then
    Begin
      Close (F);
      Exit
    End;
  {$I-} BlockWrite (F, DataIn, SizeOf(DataIn)); {$I+}
  If IOResult <> 0 then
    Begin
      Close (F);
      Exit
    End;
  {$I-} BlockWrite (F, DataIn2, SizeOf(DataIn2)); {$I+}
  If IOResult <> 0 then
    Begin
      Close (F);
      Exit
    End;
  {$I-} BlockWrite (F, DataEx^, SizeOf(DataEx^)); {$I+}
  If IOResult <> 0 then
    Begin
      Close (F);
      Exit
    End;
  Close (F);
  Error := eOk
End;

Procedure LoadMemory;
Var
  S : String;
  F : File;
Begin
  Window (27, 11, 28, 5, MsgLoadMemWinTitle, LightGray, Black);
  S := MName;
  GotoXY (35, 13);
  CursorOn (UpL, DnL);
  ReadSK (S, 12);
  CursorOff;
  If S = '' then Exit;
  MName := S;
  ErrName := MName;
  Error := eRead;
  Assign (F, MName);
  {$I-} Reset (F, 1); {$I+}
  If IOResult <> 0 then Exit;
  {$I-} BlockRead (F, S, Length(MemID)+1); {$I+}
  If IOResult <> 0 then
    Begin
      Close (F);
      Exit
    End;
  If S <> MemID then
    Begin
      Close (F);
      Exit
    End;
  {$I-} BlockRead (F, PC, SizeOf(PC)); {$I+}
  If IOResult <> 0 then
    Begin
      Close (F);
      Exit
    End;
  {$I-} BlockRead (F, DataIn, SizeOf(DataIn)); {$I+}
  If IOResult <> 0 then
    Begin
      Close (F);
      Exit
    End;
  {$I-} BlockRead (F, DataIn2, SizeOf(DataIn2)); {$I+}
  If IOResult <> 0 then
    Begin
      Close (F);
      Exit
    End;
  {$I-} BlockRead (F, DataEx^, SizeOf(DataEx^)); {$I+}
  If IOResult <> 0 then
    Begin
      Close (F);
      Exit
    End;
  Close (F);
  Error := eOk
End;

Var POldExitProc : Pointer;

{$F+}
Procedure NewExitProc; {Far;}
Begin
  { Restore old keyboard interrupt routine }
  SetIntVec (9, POldIRQ1ISR);
  ExitProc := POldExitProc;
End;
{$F-}

Procedure Init;
Begin
  POldExitProc := ExitProc;
  ExitProc := @NewExitProc;

  { Set new keyboard interrupt routine. It will tell us if
    there's a new key pressed on the keyboard. This is
    to make sure ms-windows will not stop running the debugger
    because it tries to get keyboard input from BIOS too often. }
  GetIntVec (9, POldIRQ1ISR);
  SetIntVec (9, @IRQ1ISR);
  
  If ParamCount <> 1 then
    Begin
      WriteLn (MsgProgInfo);
      WriteLn (MsgUsage[0]);
      WriteLn (MsgUsage[1]);
      WriteLn (MsgUsage[2]);
      Halt
    End;
  If MaxAvail < SizeOf(FBuf^)+(ProgMax+1)*3+DataMax+1 then
    Begin
      WriteLn (MsgOutOfMem[0], Max, MsgOutOfMem[1]);
      Halt
    End;
  ON := ParamStr(1);
  If Pos ('.', ON) <> 0 then Name := Copy (ON, 1, Pos('.', ON));
  Name := Name+'LST';
  ON := UpStr (ON);
  Name := UpStr (Name);
  GetCursorLines (UpL, DnL);
  Color (Black, LightGray);
  ClrScr;
  CursorOff;
  TextColor (Green);
  GotoXY (36, 13);
  WriteLn (MsgStartupWait);
  TextColor (LightGray);
  New (Prog);
  New (DataEx);
  New (Table);
  FillChar (Prog^, SizeOf(Prog^), 0);
  FillChar (DataEx^, SizeOf(DataEx^), 0);
  Assign (F, ON);
  {$I-} Reset (F, 1); {$I+}
  If IOResult <> 0 then
    Begin
      ClrScr;
      WriteLn (MsgBinFileNotFound[0]+ON+MsgBinFileNotFound[1]);
      CursorOn (UpL, DnL);
      Halt (1)
    End;
  If FileSize(F) > SizeOf(Prog^) then
    Begin
      Close (F);
      ClrScr;
      WriteLn (MsgBinFileSize[0], SizeOf(Prog^), MsgBinFileSize[1]);
      CursorOn (UpL, DnL);
      Halt (1)
    End;
  {$I-} BlockRead (F, Prog^, FileSize(F)); {$I+}
  If IOResult <> 0 then
    Begin
      Close (F);
      ErrName := ON;
      Error := eRead;
      OutError;
      Color (Black, LightGray);
      ClrScr;
      CursorOn (UpL, DnL);
      Halt (1)
    End;
  Close (F);
  FList := True;
  Assign (F, Name);
  {$I-} Reset (F, 1); {$I+}
  If IOResult = 0 then Close (F)
  Else FList := False;
  If FList then
    Begin
      ConvertList (Name);
      If Error <> eOk then
        Begin
          OutError;
          Color (Black, LightGray);
          ClrScr;
          CursorOn (UpL, DnL);
          Halt (1)
        End;
      MakeTable;
      If Error <> eOk then
        Begin
          OutError;
          Color (Black, LightGray);
          ClrScr;
          CursorOn (UpL, DnL);
          Halt (1)
        End;
      New (FBuf)
    End;
  ClrScr;
  MName := '';
  PCF := False;
  ClearDataIn;
  ClearDataEx;
  ClearRegs;
  RegTog := False;
  MemSel := 0;
  PageIn := 0;
  PageEx := 0;
  RegsPoint := 1;
  MemPoint := 0;
  NBreaks := 0;
  Clk := 0;
  MakeScr;
  If FList then
    Begin
      BFirst := 1;
      WFirst := 0;
      If StrNum >= BLen then {$I-} ReadBlock (1, BLen, 0) {$I+}
      Else {$I-} ReadBlock (1, StrNum, 0) {$I+};
      If Error <> eOk then
        Begin
          OutError;
          Color (Black, LightGray);
          ClrScr;
          CursorOn (UpL, DnL);
          Halt (1)
        End
    End;
  XX
End;

Begin
  { --------------------   -------------------- }
  Init;
Cycle:
  If KeyPressed then C := ReadKey
  Else Goto Cycle;
  If C <> #0 then
    Begin
      Case UpCase (C) of

         #4 : { Ctrl+D - DOS Shell }
              Begin
                Color (Black, LightGray);
                ClrScr;
                WriteLn (MsgShell);
                CursorOn (UpL, DnL);
                Exec (GetEnv('COMSPEC'), '');
                CursorOff;
                ClrScr;
                MakeScr;
                XX
              End;

        #9  : { Tab - 室  ॣࠬ }
              Begin
                RegTog := RegTog xor True;
                MakeScr;
              End;
        #13 : { Enter -  PC }
              Begin
                PCF := True;
                PCMem := PC
              End;
        #27 : { Esc - ⠭ PC }
              If PCF then
                Begin
                  PCF := False;
                  PC := PCMem;
                  If FList then PositL (Table^[PC]);
                  If Error <> eOk then
                    Begin
                      OutError;
                      Color (Black, LightGray);
                      ClrScr;
                      CursorOn (UpL, DnL);
                      Halt (1)
                    End;
                  MakeScr;
                  XX
                End;
        'C' : { C -  }
              Begin
                ClearDataIn;
                ClearDataEx;
                ClearRegs;
                MakeScr;
                If FList then PositL (Table^[PC]);
                If Error <> eOk then
                  Begin
                    OutError;
                    Color (Black, LightGray);
                    ClrScr;
                    CursorOn (UpL, DnL);
                    Halt (1)
                  End;
                XX
              End;
        #2  : { Ctrl-B -  窨 ⠭ }
              Begin
                NBreaks := 0;
                XX
              End;
        #20: { Ctrl-T -  稪 横 }
              Begin
                Clk := 0;
                MakeScr;
              End;
        'Q' : { Q - 室 }
              Goto EmEnd;
        'R' : { R - ஢ ॣ஢ }
              Begin
                CursorOn (UpL, DnL);
                PC2 := PC;
                RegsEdit;
                CursorOff;
                If PC2 <> PC then
                  Begin
                    MakeScr;
                    If FList then PositL (Table^[PC]);
                    If Error <> eOk then
                      Begin
                        OutError;
                        Color (Black, LightGray);
                        ClrScr;
                        CursorOn (UpL, DnL);
                        Halt (1)
                      End;
                    XX
                  End;
              End;
        'M' : { M - ஢  }
              Begin
                CursorOn (UpL, DnL);
                MemEdit;
                CursorOff
              End;
        'N' : { 롮 뢠  }
              Begin
                MemSel := 1-MemSel;
                MakeMem
              End;
        '0'..'9' : {  ࠭ 譥  }
              Begin
                MemSel := 1;
                PageEx := Ord (C) - 48;
                MakeMem
              End;
        'S' : {  蠣 }
              Begin
                OneStep;
                If FList then PositL (Table^[PC]);
                If Error <> eOk then
                  Begin
                    OutError;
                    Color (Black, LightGray);
                    ClrScr;
                    CursorOn (UpL, DnL);
                    Halt (1)
                  End;
                MakeScr;
                XX
              End;
        'P' : { 믮 楤 }
              Begin
                Proc;
                If FList then PositL (Table^[PC]);
                If Error <> eOk then
                  Begin
                    OutError;
                    Color (Black, LightGray);
                    ClrScr;
                    CursorOn (UpL, DnL);
                    Halt (1)
                  End;
                MakeScr;
                XX
              End;
        'G' : { 믮  窨 ⠭ }
              Begin
                Go;
                If Flist then PositL (Table^[PC]);
                If Error <> eOk then
                  Begin
                    OutError;
                    Color (Black, LightGray);
                    ClrScr;
                    CursorOn (UpL, DnL);
                    Halt (1)
                  End;
                MakeScr;
                XX
              End;
        'B' : { ⠭/  ⠭ }
              If Flist then
                Begin
                  If FBuf^[WFirst][13] <> ' ' then
                    Begin
                      Breaks (Valu('$'+Copy(FBuf^[WFirst],8,4)));
                      XX
                    End
                End
              Else
                Begin
                  Breaks (PC);
                  XX
                End;
        'A' : {  ⨭  ⥪饣 PC }
              If FList then
                Begin
                  PositL (Table^[PC]);
                  If Error <> eOk then
                    Begin
                      OutError;
                      Color (Black, LightGray);
                      ClrScr;
                      CursorOn (UpL, DnL);
                      Halt (1)
                    End;
                  XX
                End;
        'W' : {  ࠭ 譥  }
              If MemSel = 1 then
                Begin
                  WinAddrEx;
                End;
      End
    End
  Else
    Begin
      C := ReadKey;
      Case C of
        #81 : { PgDn - ⨭ }
              Begin
                For I := 1 to WH do Down;
                If Error <> eOk then
                  Begin
                    OutError;
                    Color (Black, LightGray);
                    ClrScr;
                    CursorOn (UpL, DnL);
                    Halt (1)
                  End;
                If FList and (Mem[0:1047] and 3 <> 0) then
                  Begin
                    PC := Valu('$'+Copy(FBuf^[WFirst],8,4));
                    MakeScr
                  End
                Else If not FList then MakeScr;
                XX
              End;
        #73 : { PgUp - ⨭ }
              Begin
                For I := 1 to Wh do Up;
                If Error <> eOk then
                  Begin
                    OutError;
                    Color (Black, LightGray);
                    ClrScr;
                    CursorOn (UpL, DnL);
                    Halt (1)
                  End;
                If FList and (Mem[0:1047] and 3 <> 0) then
                  Begin
                    PC := Valu('$'+Copy(FBuf^[WFirst],8,4));
                    MakeScr
                  End
                Else If not FList then MakeScr;
                XX
              End;
        #119: { Ctrl-Home - 砫 ⨭ }
              If FList then
                Begin
                  If StrNum >= BLen then ReadBlock (1, BLen, 0)
                  Else ReadBlock (1, StrNum, 0);
                  If Error <> eOk then
                    Begin
                      OutError;
                      Color (Black, LightGray);
                      ClrScr;
                      CursorOn (UpL, DnL);
                      Halt (1)
                    End;
                  BFirst := 1;
                  WFirst := 0;
                  XX
                End;
        #117: { Ctrl-End -  ⨭ }
              If FList and (StrNum > WH) then
                Begin
                  BFirst := (BLen-WH) * (StrNum div (BLen-WH))+1;
                  ReadBlock (BFirst, BLen, 0);
                  If Error <> eOk then
                    Begin
                      OutError;
                      Color (Black, LightGray);
                      ClrScr;
                      CursorOn (UpL, DnL);
                      Halt (1)
                    End;
                  WFirst := StrNum-BFirst-WH+1;
                  XX
                End;
        #71 : { Home -  }
              Begin
                MemSel := 0;
                PageIn := 0;
                MakeMem
              End;
        #79 : { End -  }
              Begin
                MemSel := 0;
                PageIn := 1;
                MakeMem
              End;
        #132: { Ctrl-PgUp -  }
              Begin
                If (MemSel = 0) and (PageIn > 0) then Dec (PageIn)
                Else
                  If (MemSel = 1) and (PageEx > 0) then Dec (PageEx);
                MakeMem
              End;
        #118: { Ctrl-PgDn -  }
              Begin
                If (MemSel = 0) and (PageIn < 2) then Inc (PageIn)
                Else
                  If (MemSel = 1) and (PageEx < DataMax div 128) then Inc (PageEx);
                MakeMem
              End;
        #72 : { Up - ⨭ }
              Begin
                Up;
                If Error <> eOk then
                  Begin
                    OutError;
                    Color (Black, LightGray);
                    ClrScr;
                    CursorOn (UpL, DnL);
                    Halt (1)
                  End;
                If FList and (Mem[0:1047] and 3 <> 0) then
                  Begin
                    PC := Valu('$'+Copy(FBuf^[WFirst],8,4));
                    MakeScr
                  End
                Else If not FList then MakeScr;
                XX
              End;
        #80 : { Down - ⨭ }
              Begin
                Down;
                If Error <> eOk then
                  Begin
                    OutError;
                    Color (Black, LightGray);
                    ClrScr;
                    CursorOn (UpL, DnL);
                    Halt (1)
                  End;
                If FList and (Mem[0:1047] and 3 <> 0) then
                  Begin
                    PC := Valu('$'+Copy(FBuf^[WFirst],8,4));
                    MakeScr
                  End
                Else If not FList then MakeScr;
                XX
              End;
        #59 : { F1 -  }
              Begin
                Help;
                MakeScr;
                XX
              End;
        #60 : { F2 -    䠩 }
              Begin
                SaveMemory;
                If Error <> eOk then
                  Begin
                    OutError;
                    Error := eOk
                  End;
                Color (Black, Green);
                ClrScr;
                MakeScr;
                XX
              End;
        #61 : { F3 - 㧨   䠩 }
              Begin
                LoadMemory;
                If Error <> eOk then
                  Begin
                    OutError;
                    Error := eOk
                  End;
                Color (Black, Green);
                ClrScr;
                MakeScr;
                If FList then PositL (Table^[PC]);
                If Error <> eOk then
                  Begin
                    OutError;
                    Color (Black, LightGray);
                    ClrScr;
                    CursorOn (UpL, DnL);
                    Halt (1)
                  End;
                XX
              End;
        #62 : { F4 -  }
              Begin
                Calc (MakeScrXX);
                MakeScrXX
              End
      End
    End;
  Goto Cycle;
EmEnd:
  If FList then
    Begin
      Dispose (FBuf);
      Assign (F, LN);
      Erase (F)
    End;
  Dispose (Table);
  Dispose (DataEx);
  Dispose (Prog);
  Color (Black, LightGray);
  ClrScr;
  WriteLn (MsgProgInfo);
  CursorOn (UpL, DnL)
End.
