  {----------------------------------------------------------------------}
  { Project title:       EMF52L                                          }
  { Project description: Software debugger for C51/C52 micro controllers }
  { Unit:                EmMCS52                                         }
  { Unit Description:    Main MCU emulation unit (emulation, disassembly,}
  {                      user interface, etc)                            }
  { Author:              Alexei A. Frounze (c) 2003                      }
  { Version and date:    2.30 april 2003                                 }
  { History:                                                             }
  {  2000? 2001? 2002?   Bugfixes in AJMP and ACALL: address calculation }
  {  2001-10-13 (oct 13) Bugfixed in ADDC and SUBB: PSW flags            }
  {  2003-02-02 (feb 2)  Clock counter added                             }
  {  2003-04-06 (apr 6)  added a workaround to the ms-windows problem:   }
  {                      windows could stop the debugger operation when  }
  {                      running a time-consuming piece of code, for     }
  {                      example a 5-seconds-long delay subroutine,      }
  {                      initiated by pressing P or G. Visually this     }
  {                      appeared as if the code took ifinite time to    }
  {                      complete and it could only be interrupted by    }
  {                      the space bar key. This happened because the    }
  {                      debugger asked BIOS if there's a keystroke ready}
  {                      too often. Windows decided that the debugger is }
  {                      an interactive application busy-wating on the   }
  {                      keyboard and paused execution of the debugger   }
  {                      until a key was pressed on the keyboard.        }
  {                      So, the code could be forced to complete by     }
  {                      pressing and releasing keys on the keyboard,    }
  {                      for example, the Ctrl key.                      }
  {----------------------------------------------------------------------}

{$IFDEF VER55}
  {$R-}
{$ELSE}
  {$Q-,R-}
{$ENDIF}

  {------------------------------------------------------------------------}
                                Unit EmMCS52;
  {------------------------------------------------------------------------}


  {------------------------------------------------------------------------}
                                  Interface
  {------------------------------------------------------------------------}

  {------------------------ 㯭 ⠭ ---------------------------}

Const
{  ⥬}
  HexaDec : Set of Char = ['0'..'9', 'A'..'F'];
  Nums    : Set of Char = ['0'..'9'];

{   }
  ProgMax = 32767;
  DataMax = 65534;

{  ⠭ }
  CY  = $80;
  AC  = $40;
  F0  = $20;
  RS1 = $10;
  RS0 = $08;
  OV  = $04;
  F1  = $02;
  P   = $01;

{  ॣ஢ ᯥ樠 㭪権 }
  P0  = $80;
  SP  = $81;
  DPL = $82;
  DPH = $83;
  PCON= $87;
  TCON= $88;
  TMOD= $89;
  TL0 = $8A;
  TL1 = $8B;
  TH0 = $8C;
  TH1 = $8D;
  P1  = $90;
  SCON= $98;
  SBUF= $99;
  P2  = $A0;
  IE  = $A8;
  SADR= $A9;
  P3  = $B0;
  IPH = $B7;
  IP  = $B8;
  SADN= $B9;
  T2CN= $C8;
  T2MD= $C9;
  RCPL= $CA;
  RCPH= $CB;
  TL2 = $CC;
  TH2 = $CD;
  PSW = $D0;
  ACC = $E0;
  B   = $F0;

  {--------------------------  - ᨢ  -----------------------}

Type
  ProgArr   = Array [0..ProgMax] of Byte;
  DataExArr = Array [0..DataMax] of Byte;

  {----------------------- 㯭 ६ ---------------------------}

Var
  Prog   : ^ProgArr;
  DataEx : ^DataExArr;
  DataIn : Array   [0..255]   of Byte;
  DataIn2: Array  [128..255]  of Byte;
  BP     : Array   [1 ..16]   of Word;
  NBreaks,
  MemSel,
  PageIn,
  RegsPoint,
  MemPoint: Byte;
  RegTog  : Boolean;
  W, PageEx,
  PC, PC2 : Word;
  Clk     : LongInt;


  {--------------------- 㯭 楤  㭪樨 --------------------}

  Function Hex (B : Byte) : String;
  Function JR (PC : Word; R : Byte) : Word;
  Function BitOn (Bit : Byte) : Boolean;
  Procedure SetBit (Bit : Byte);
  Procedure ClrBit (Bit : Byte);
  Procedure CplBit (Bit : Byte);
  Procedure ClearProg;
  Procedure ClearDataIn;
  Procedure ClearDataEx;
  Procedure ClearRegs;
  Procedure Add (B : Byte);
  Procedure AddC (B : Byte);
  Procedure SubB (B : Byte);
  Procedure DaA;
  Procedure DivAB;
  Procedure MulAB;
  Procedure OneStep;
  Function Deci (Var H : String) : Word;
  Procedure MakeMem;
  Procedure MakeScr;
  Procedure RegsEdit;
  Procedure MemEdit;
  Procedure Proc;
  Procedure Go;
  Procedure WinAddrEx;

{-- 頥 ⥯ 16- --}
Function S16 (N : Byte) : Word;

{-- 頥 . ப -  ⭮⥫쭮  室 --}
Function JRel (PC : Word; R : Byte) : String;

{-- 頥 ப -    , ᫥ 㢥. --}
Function Step (Var PC : Word) : String;

Var POldIRQ1ISR : Pointer;
Var HasIRQ1Fired : Boolean;
Procedure IRQ1ISR; Interrupt;

  {------------------------------------------------------------------------}
                                Implementation
  {------------------------------------------------------------------------}

Uses
  CRT,
  ProfText,
  S_Input,
  EMF52L_L;

Var
  UpL, DnL : Byte;


  {------------------------------------- ⠭ ------------------------}
Const
  OneByte   : Set of Byte = [         {⮢ 樨}
  $00, $03, $04, $06..$0F, $13, $14, $16..$1F, $22, $23, $26..$2F, $32, $33,
  $36..$3F, $46..$4F, $56..$5F, $66..$6F, $73, $83, $84, $93, $96..$9F, $A3,
  $A4, $A5, $B3, $C3, $C4, $C6..$CF, $D3, $D4, $D6, $D7, $E0, $E2..$E4,
  $E6..$EF, $F0, $F2..$F4, $F6..$FF];

  TwoByte   : Set of Byte = [         {塠⮢ 樨}
  $01, $05, $11, $15, $21, $24, $25, $31, $34, $35, $40..$42, $44, $45,
  $50..$52, $54, $55, $60..$62, $64, $65, $70..$72, $74, $76..$7F, $80..$82,
  $86..$8F, $91, $92, $94, $95, $A0..$A2, $A6..$AF, $B0..$B2, $C0..$C2, $C5,
  $D0..$D2, $D8..$DF, $E1, $E5, $F1, $F5];

  ThreeByte : Set of Byte = [         {塠⮢ 樨}
  $02, $10, $12, $20, $30, $43, $53, $63, $75, $85, $90, $B4..$BF, $D5];

  {--------------------------   樨 ----------------------------}

  CAJMP     : Set of Byte = [
  $01, $21, $41, $61, $81, $A1, $C1, $E1];

  CACALL    : Set of Byte = [
  $11, $31, $51, $71, $91, $B1, $D1, $F1];

  CMOV      : Set of Byte = [
  $74..$7F, $85..$8F, $90, $92, $A2, $A6..$AF, $E5..$EF, $F5..$FF];

  CINC      : Set of Byte = [
  $04..$0F, $A3];

  CDEC      : Set of Byte = [
  $14..$1F];

  CADD      : Set of Byte = [
  $24..$2F];

  CADDC     : Set of Byte = [
  $34..$3F];

  CORL      : Set of Byte = [
  $42..$4F, $72, $A0, $B0];

  CANL      : Set of Byte = [
  $52..$5F, $82];

  CXRL      : Set of Byte = [
  $62..$6F];

  CSUBB     : Set of Byte = [
  $94..$9F];

  CCJNE     : Set of Byte = [
  $B4..$BF];

  CDJNZ     : Set of Byte = [
  $D5, $D8..$DF];

  CXCH      : Set of Byte = [
  $C5..$CF];

  CXCHD     : Set of Byte = [
  $D6, $D7];

  CCLR      : Set of Byte = [
  $C2, $C3, $E4];

  CCPL      : Set of Byte = [
  $B2, $B3, $F4];

  CSETB     : Set of Byte = [
  $D2, $D3];

  MOVX      : Set of Byte = [
  $E0, $E2, $E3, $F0, $F2, $F3];

  {---------------- 樨, ।  ࠭ ---------------}

  Bit_Rel   : Set of Byte = [
  $10, $20, $30];

  Rel       : Set of Byte = [
  $40, $50, $60, $70, $80];

  C_Bit     : Set of Byte = [
  $72, $82, $A2];

  C_N_Bit   : Set of Byte = [
  $A0, $B0];

  Bit       : Set of Byte = [
  $B2, $C2, $D2];

  Ad        : Set of Byte = [
  $05, $15, $C0, $D0];

  Ad_A      : Set of Byte = [
  $42, $52, $62, $F5];

  A_Ad      : Set of Byte = [
  $25, $35, $45, $55, $65, $95, $C5, $E5];

  Ad16      : Set of Byte = [
  $02, $12];

  Ad_D      : Set of Byte = [
  $43, $53, $63, $75];

  Carry     : Set of Byte = [
  $B3, $C3, $D3];

  Accum     : Set of Byte = [
  $04, $14, $E4, $F4];

  A_D       : Set of Byte = [
  $24, $34, $44, $54, $64, $74, $94];

  AdR0      : Set of Byte = [
  $06, $16];

  AdR1      : Set of Byte = [
  $07, $17];

  A_AdR0    : Set of Byte = [
  $26, $36, $46, $56, $66, $96, $C6, $D6, $E6];

  A_AdR1    : Set of Byte = [
  $27, $37, $47, $57, $67, $97, $C7, $D7, $E7];

  A_R0      : Set of Byte = [
  $28, $38, $48, $58, $68, $98, $C8, $E8];

  A_R1      : Set of Byte = [
  $29, $39, $49, $59, $69, $99, $C9, $E9];

  A_R2      : Set of Byte = [
  $2A, $3A, $4A, $5A, $6A, $9A, $CA, $EA];

  A_R3      : Set of Byte = [
  $2B, $3B, $4B, $5B, $6B, $9B, $CB, $EB];

  A_R4      : Set of Byte = [
  $2C, $3C, $4C, $5C, $6C, $9C, $CC, $EC];

  A_R5      : Set of Byte = [
  $2D, $3D, $4D, $5D, $6D, $9D, $CD, $ED];

  A_R6      : Set of Byte = [
  $2E, $3E, $4E, $5E, $6E, $9E, $CE, $EE];

  A_R7      : Set of Byte = [
  $2F, $3F, $4F, $5F, $6F, $9F, $CF, $EF];

  RR0       : Set of Byte = [
  $08, $18];

  RR1       : Set of Byte = [
  $09, $19];

  RR2       : Set of Byte = [
  $0A, $1A];

  RR3       : Set of Byte = [
  $0B, $1B];

  RR4       : Set of Byte = [
  $0C, $1C];

  RR5       : Set of Byte = [
  $0D, $1D];

  RR6       : Set of Byte = [
  $0E, $1E];

  RR7       : Set of Byte = [
  $0F, $1F];

  {----------------- ६ 믮 権  横 -----------------}

  ClkIncrement : Array[Byte] of Byte = (
    {x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 xA xB xC xD xE xF}
{0x}  1, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
{1x}  2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
{2x}  2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
{3x}  2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
{4x}  2, 2, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
{5x}  2, 2, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
{6x}  2, 2, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
{7x}  2, 2, 2, 2, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
{8x}  2, 2, 2, 2, 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
{9x}  2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
{Ax}  2, 2, 1, 2, 4, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
{Bx}  2, 2, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
{Cx}  2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
{Dx}  2, 2, 1, 1, 1, 2, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2,
{Ex}  2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
{Fx}  2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1);

Procedure Color (B, T : Byte);
Begin
  TextColor (T);
  TextBackGround (B)
End;

{-- 頥 ⥯ 16- --}
Function S16 (N : Byte) : Word;
Var
  Z,
  I : Word;
Begin
  Z := 1;
  If N > 0 then
    For I := 1 to N do
      Z := Z * 16;
  S16 := Z
End;

Function Hex (B : Byte) : String;
Var
  S : String;
  H,
  L : Byte;
Begin
  H := B shr 4;
  L := B and 15;
  If H < 10 then S := Chr (48 + H)
  Else S := Chr (55 + H);
  If L < 10 then S := S + Chr (48 + L)
  Else S := S + Chr (55 + L);
  Hex := S
End;

Function JR (PC : Word; R : Byte) : Word;
Var
  A : Word;
Begin
  If R < $80 then A := PC+R
  Else A := PC+R-256;
  JR := A
End;

Function BitOn (Bit : Byte) : Boolean;
Var P : Boolean;
Begin
  Case Bit of
    $00..$7F :
      P := DataIn [$20 + Bit div 8] and (1 shl (Bit mod 8)) <> 0;
    $80..$F7 :
      If Bit and $0F < $08 then
        P := DataIn [Bit and $F0] and (1 shl (Bit mod 8)) <> 0
      Else
        P := DataIn [Bit and $F0 + $08] and (1 shl (Bit mod 8)) <> 0
  End;
  BitOn := P
End;

Procedure SetBit (Bit : Byte);
Begin
  Case Bit of
    $00..$7F :
      DataIn [$20 + Bit div 8] :=
                        DataIn [$20 + Bit div 8] or (1 shl (Bit mod 8));
    $80..$F7 :
      If Bit and $0F < $08 then
        DataIn [Bit and $F0] := DataIn [Bit and $F0] or
                                                     (1 shl (Bit mod 8))
      Else
        DataIn [Bit and $F0 + $08] := DataIn [Bit and $F0 + $08] or
                                                     (1 shl (Bit mod 8))
  End
End;

Procedure ClrBit (Bit : Byte);
Begin
  Case Bit of
    $00..$7F :
      DataIn [$20 + Bit div 8] := DataIn [$20 + Bit div 8] and ($FF - (1 shl (Bit mod 8)));
    $80..$F7 :
      If Bit and $0F < $08 then
        DataIn [Bit and $F0] := DataIn [Bit and $F0] and ($FF - (1 shl (Bit mod 8)))
      Else
        DataIn [Bit and $F0 + $08] := DataIn [Bit and $F0 + $08] and ($FF - (1 shl (Bit mod 8)))
  End
End;

Procedure CplBit (Bit : Byte);
Begin
  If BitOn (Bit) then ClrBit (Bit)
  Else SetBit (Bit)
End;

Procedure ClearProg;
Var I : Word;
Begin
  For I := 0 to ProgMax do Prog^[I] := 0
End;

Procedure ClearDataIn;
Var I : Word;
Begin
  For I := 0 to 255 do DataIn [I] := 0;
  For I := 128 to 255 do DataIn2[I] := 0
End;

Procedure ClearDataEx;
Var I : Word;
Begin
  For I := 0 to DataMax do DataEx^[I] := 0
End;

Procedure ClearRegs;
Var I : Byte;
Begin
  For I := 0 to 255 do
    DataIn[I] := 0;
  DataIn[SP] := 7;
  DataIn[P0] := $FF;
  DataIn[P1] := $FF;
  DataIn[P2] := $FF;
  DataIn[P3] := $FF;
  PC := 0
End;

Function Reg (N : Byte) : Byte;
Var
  Bank : Byte;
Begin
  Bank := (DataIn[PSW] shr 3) and 3;
  Reg := DataIn [Bank*8+N]
End;

Procedure SetReg (N, B : Byte);
Var
  Bank : Byte;
Begin
  Bank := (DataIn[PSW] shr 3) and 3;
  DataIn [Bank*8+N] := B
End;

Procedure IncReg (N : Byte);
Var
  Bank : Byte;
Begin
  Bank := (DataIn[PSW] shr 3) and 3;
  Inc (DataIn [Bank*8+N])
End;

Procedure DecReg (N : Byte);
Var
  Bank : Byte;
Begin
  Bank := (DataIn[PSW] shr 3) and 3;
  Dec (DataIn [Bank*8+N])
End;

Procedure Add (B : Byte);
Var
  FCY,
  FAC,
  FOV: Boolean;
Begin
  FCY := (DataIn[ACC] + B) > $FF;
  FAC := ((DataIn[ACC] and $0F) + (B and $0F)) > $0F;
  FOV := ((((DataIn[ACC] and $7F) + (B and $7F)) >  $7F) and not FCY) or
        ((((DataIn[ACC] and $7F) + (B and $7F)) <= $7F) and     FCY);
  DataIn[ACC] := DataIn[ACC] + B;
  DataIn[PSW] := (DataIn[PSW] and not (CY+AC+OV)) or
         (CY*Ord(FCY) + AC*Ord(FAC) + OV*Ord(FOV))
End;

Procedure AddC (B : Byte);
Var
  C : Byte;
  FCY,
  FAC,
  FOV: Boolean;
Begin
  C := Ord (DataIn[PSW] and CY = CY);
  FCY := (DataIn[ACC] + B + C) > $FF;
  FAC := ((DataIn[ACC] and $0F) + (B and $0F) + C) > $0F;
  FOV := ((((DataIn[ACC] and $7F) + (B and $7F) + C) >  $7F) and not FCY) or
        ((((DataIn[ACC] and $7F) + (B and $7F) + C) <= $7F) and     FCY);
  DataIn[ACC] := DataIn[ACC] + B + C;
  DataIn[PSW] := (DataIn[PSW] and not(CY+AC+OV)) or
         (CY*Ord(FCY) + AC*Ord(FAC) + OV*Ord(FOV))
End;

Procedure SubB (B : Byte);
Var
  C : Byte;
  FCY,
  FAC,
  FOV: Boolean;
Begin
  C := Ord (DataIn[PSW] and CY = CY);
  FCY := DataIn[ACC] < B + C;
  FAC := (DataIn[ACC] and $0F) < (B and $0F) + C;
  FOV := (((DataIn[ACC] and $7F) <  (B and $7F) + C) and not FCY) or
         (((DataIn[ACC] and $7F) >= (B and $7F) + C) and     FCY);
  DataIn[ACC] := DataIn[ACC] - B - C;
  DataIn[PSW] := (DataIn[PSW] and not(CY+AC+OV)) or
         (CY*Ord(FCY) + AC*Ord(FAC) + OV*Ord(FOV))
End;

Procedure DaA;
Begin
  If ((DataIn[Acc] and $0F) > 9) or (DataIn[PSW] and AC = AC) then
    Begin
      If DataIn[Acc]+$06 > $FF then DataIn[PSW] := DataIn[PSW] or CY;
      DataIn[Acc] := DataIn[Acc]+$06
    End;
  If ((DataIn[Acc] and $F0) > $90) or (DataIn[PSW] and CY = CY) then
    Begin
      If DataIn[Acc]+$60 > $FF then DataIn[PSW] := DataIn[PSW] or CY;
      DataIn[Acc] := DataIn[Acc]+$60
    End
End;

Procedure DivAB;
Var
  C : Byte;
Begin
  If DataIn[B] = 0 then
    Begin
      DataIn[Acc] := Random (256);
      DataIn[B  ] := Random (256);
      DataIn[PSW] := DataIn[PSW] and ($FF-CY);
      DataIn[PSW] := DataIn[PSW] or  OV
    End
  Else
    Begin
      DataIn[PSW] := DataIn[PSW] and ($FF-CY-OV);
      C   := DataIn[B];
      DataIn[B] := DataIn[Acc] mod C;
      DataIn[Acc] := DataIn[Acc] div C
    End
End;

Procedure MulAB;
Var
  C : Word;
Begin
  C := DataIn[Acc]*DataIn[B];
  If C > $FF then DataIn[PSW] := DataIn[PSW] or OV;
  DataIn[PSW] := DataIn[PSW] and ($FF-CY);
  DataIn[B] := Hi (C);
  DataIn[Acc] := Lo (C)
End;

Procedure OneStep;
Var
  K : Word;
  B1, B2, B3,
  I, J : Byte;
  Temp : Word;
Begin
  If PC > ProgMax then PC := PC mod (ProgMax+1);
  B1 := Prog^[PC];

  If PC+1 <= ProgMax then B2 := Prog^[PC+1]
  Else B2 := 0;
  
  If PC+2 <= ProgMax then B3 := Prog^[PC+2]
  Else B3 := 0;

  Case B1 of
    {--------------------------- MOV --------------------------}
    $74 : {MOV A,#d}
    Begin DataIn[ACC] := B2; Inc (PC, 2) End;

    $75 : {MOV ad,#d}
    Begin DataIn [B2] := B3; Inc (PC, 3) End;

    $76, $77 : {MOV @Ri,#d}
    Begin
      If Reg(B1-$76) < $80 then
        DataIn [Reg(B1-$76)] := B2
      Else
        DataIn2[Reg(B1-$76)] := B2;
      Inc (PC, 2)
    End;

    $78..$7F : {MOV Rn,#d}
      Begin
        SetReg (B1-$78, B2);
        Inc (PC, 2)
      End;

    $83 : {MOVC A,@A+PC}
    Begin Inc (PC); DataIn[Acc] := Prog^[DataIn[Acc]+PC] End;

    $85 : {MOV add,ads}
    Begin DataIn [B3] := DataIn [B2]; Inc (PC, 3) End;

    $86, $87 : {MOV ad,@Ri}
    Begin
      If Reg(B1-$86) < $80 then
        DataIn [B2] := DataIn [Reg(B1-$86)]
      Else
        DataIn [B2] := DataIn2[Reg(B1-$86)];
      Inc (PC, 2)
    End;

    $88..$8F : {MOV ad,Rn}
      Begin
        DataIn [B2] := Reg(B1-$88);
        Inc (PC, 2)
      End;

    $90 : {MOV DPTR,#d16}
    Begin
      DataIn[DPL] := B3;
      DataIn[DPH] := B2;
      Inc (PC, 3)
    End;

    $93 : {MOVC A,@A+DPTR}
    Begin
      DataIn[Acc] := Prog^[DataIn[Acc]+DataIn[DPH]*256+DataIn[DPL]];
      Inc (PC)
    End;

    $A6, $A7 : {MOV @Ri,ad}
    Begin
      If Reg(B1-$A6) < $80 then
        DataIn [Reg(B1-$A6)] := DataIn [B2]
      Else
        DataIn2[Reg(B1-$A6)] := DataIn [B2];
      Inc (PC, 2)
    End;

    $A8..$AF : {MOV Rn,ad}
    Begin
      SetReg(B1-$A8, DataIn [B2]);
      Inc (PC, 2)
    End;

    $E0 : {MOVX A,@DPTR}
    Begin
      DataIn[Acc] := DataEx^[DataIn[DPH]*256+DataIn[DPL]];
      Inc (PC)
    End;

    $E2, $E3 : {MOVX A,@Ri}
    Begin DataIn[Acc] := DataEx^[Reg(B1-$E2)]; Inc (PC) End;

    $E5 : {MOV A,ad}
    Begin DataIn[ACC] := DataIn [B2]; Inc (PC, 2) End;

    $E6, $E7 : {MOV A,@Ri}
    Begin
      If Reg(B1-$E6) < $80 then
        DataIn[ACC] := DataIn [Reg(B1-$E6)]
      Else
        DataIn[ACC] := DataIn2[Reg(B1-$E6)];
      Inc (PC)
    End;

    $E8..$EF : {MOV A,Rn}
    Begin DataIn[ACC] := Reg(B1-$E8); Inc (PC) End;

    $F0 : {MOVX @DPTR,A}
    Begin
      DataEx^[DataIn[DPL]+DataIn[DPH]*256] := DataIn[Acc];
      Inc (PC)
    End;

    $F2, $F3 : {MOVX @Ri,A}
    Begin DataEx^[Reg(B1-$F2)] := DataIn[Acc]; Inc (PC) End;

    $F5 : {MOV ad,A}
    Begin DataIn [B2] := DataIn [ACC]; Inc (PC, 2) End;

    $F6, $F7 : {MOV @Ri,A}
    Begin
      If Reg(B1-$F6) < $80 then
        DataIn [Reg(B1-$F6)] := DataIn[ACC]
      Else
        DataIn2[Reg(B1-$F6)] := DataIn[ACC];
      Inc (PC)
    End;

    $F8..$FF : {MOV Rn,A}
    Begin SetReg(B1-$F8, DataIn[ACC]); Inc (PC) End;

    {--------------------------- INC --------------------------}
    $04 : {INC A}
    Begin Inc (DataIn[Acc]); Inc (PC) End;

    $05 : {INC ad}
    Begin Inc (DataIn [B2]); Inc (PC, 2) End;

    $06, $07 : {INC @Ri}
    Begin
      If Reg(B1-$06) < $80 then
        Inc (DataIn [Reg(B1-$06)])
      Else
        Inc (DataIn2[Reg(B1-$06)]);
      Inc (PC)
    End;

    $08..$0F : {INC Rn}
    Begin IncReg(B1-$08); Inc (PC) End;

    $A3 : {INC DPTR}
    Begin
      W := DataIn[DPL]+256*DataIn[DPH];
      Inc (W);
      DataIn[DPL] := Lo (W);
      DataIn[DPH] := Hi (W);
      Inc (PC)
    End;

    {--------------------------- DEC --------------------------}
    $14 : {DEC A}
    Begin Dec (DataIn[Acc]); Inc (PC) End;

    $15 : {DEC ad}
    Begin Dec (DataIn [B2]); Inc (PC, 2) End;

    $16, $17 : {DEC @Ri}
    Begin
      If Reg(B1-$16) < $80 then
        Dec (DataIn [Reg(B1-$16)])
      Else
        Dec (DataIn2[Reg(B1-$16)]);
      Inc (PC)
    End;

    $18..$1F : {DEC Rn}
    Begin DecReg(B1-$18); Inc (PC) End;

    {--------------------------- ADD --------------------------}
    $24 : {ADD A,#d}
    Begin Add (B2); Inc (PC, 2) End;

    $25 : {ADD A,ad}
    Begin Add (DataIn [B2]); Inc (PC, 2) End;

    $26, $27 : {ADD A,@Ri}
    Begin
      If Reg(B1-$26) < $80 then
        Add (DataIn [Reg(B1-$26)])
      Else
        Add (DataIn2[Reg(B1-$26)]);
      Inc (PC)
    End;

    $28..$2F : {ADD A,Rn}
    Begin Add (Reg(B1-$28)); Inc (PC) End;

    {--------------------------- ADDC -------------------------}
    $34 : {ADDC A,#d}
    Begin AddC (B2); Inc (PC, 2) End;

    $35 : {ADDC A,ad}
    Begin AddC (DataIn [B2]); Inc (PC, 2) End;

    $36, $37 : {ADDC A,@Ri}
    Begin
      If Reg(B1-$36) < $80 then
        AddC (DataIn [Reg(B1-$36)])
      Else
        AddC (DataIn2[Reg(B1-$36)]);
      Inc (PC)
    End;

    $38..$3F : {ADDC A,Rn}
    Begin AddC (Reg(B1-$38)); Inc (PC) End;

    {--------------------------- SUBB -------------------------}
    $94 : {SUBB A,#d}
    Begin SubB (B2); Inc (PC, 2) End;

    $95 : {SUBB A,ad}
    Begin SubB (DataIn [B2]); Inc (PC, 2) End;

    $96, $97 : {SUBB A,@Ri}
    Begin
      If Reg(B1-$96) < $80 then
        SubB (DataIn [Reg(B1-$96)])
      Else
        SubB (DataIn2[Reg(B1-$96)]);
      Inc (PC)
    End;

    $98..$9F : {SUBB A,Rn}
    Begin SubB (Reg(B1-$98)); Inc (PC) End;

    {--------------------------- DA A -------------------------}
    $D4 : {DA A}
    Begin DaA; Inc (PC) End;

    {-------------------------- DIV AB ------------------------}
    $84 : {DIV AB}
    Begin DivAB; Inc (PC) End;

    {-------------------------- MUL AB ------------------------}
    $A4 : {MUL AB}
    Begin MulAB; Inc (PC) End;

    {--------------------------- RR A -------------------------}
    $03 : {RR A}
    Begin
      DataIn[Acc] := $80 * (DataIn[Acc] and $01) + DataIn[Acc] shr 1;
      Inc (PC)
    End;

    {--------------------------- RL A -------------------------}
    $23 : {RL A}
    Begin
      DataIn[Acc] := (DataIn[Acc] and $80) div $80 + DataIn[Acc] shl 1;
      Inc (PC)
    End;

    {-------------------------- RRC A -------------------------}
    $13 : {RRC A}
    Begin
      I := DataIn[PSW] and CY;
      DataIn[PSW] := (DataIn[PSW] and $7F) or CY*(DataIn[Acc] and $01);
      DataIn[Acc] := I + DataIn[Acc] shr 1;
      Inc (PC)
    End;

    {-------------------------- RLC A -------------------------}
    $33 : {RLC A}
    Begin
      I := (DataIn[PSW] and CY) div CY;
      DataIn[PSW] := (DataIn[PSW] and $7F) or (DataIn[Acc] and $80);
      DataIn[Acc] := I + DataIn[Acc] shl 1;
      Inc (PC)
    End;

    {-------------------------- CLR A -------------------------}
    $E4 : {CLR A}
    Begin DataIn[Acc] := 0; Inc (PC) End;

    {-------------------------- CPL A -------------------------}
    $F4 : {CPL A}
    Begin DataIn[Acc] := DataIn[Acc] xor $FF; Inc (PC) End;

    {------------------------- SWAP A -------------------------}
    $C4 : {SWAP A}
    Begin
      DataIn[Acc] := DataIn[Acc] shl 4 + DataIn[Acc] Shr 4;
      Inc (PC)
    End;

    {--------------------------- ORL --------------------------}
    $42 : {ORL ad,A}
    Begin DataIn[B2] := DataIn[B2] or DataIn[Acc]; Inc (PC, 2) End;

    $43 : {ORL ad,#d}
    Begin DataIn[B2] := DataIn[B2] or B3; Inc (PC, 3) End;

    $44 : {ORL A,#d}
    Begin DataIn[ACC] := DataIn[ACC] or B2; Inc (PC, 2) End;

    $45 : {ORL A,ad}
    Begin DataIn[ACC] := DataIn[ACC] or DataIn [B2]; Inc (PC, 2) End;

    $46, $47 : {ORL A,@Ri}
    Begin
      If Reg(B1-$46) < $80 then
        DataIn[ACC] := DataIn[ACC] or DataIn [Reg(B1-$46)]
      Else
        DataIn[ACC] := DataIn[ACC] or DataIn2[Reg(B1-$46)];
      Inc (PC)
    End;

    $48..$4F : {ORL A,Rn}
    Begin DataIn[ACC] := DataIn[ACC] or Reg(B1-$48); Inc (PC) End;

    {--------------------------- ANL --------------------------}
    $52 : {ANL ad,A}
    Begin DataIn[B2] := DataIn[B2] and DataIn[Acc]; Inc (PC, 2) End;

    $53 : {ANL ad,#d}
    Begin DataIn[B2] := DataIn[B2] and B3; Inc (PC, 3) End;

    $54 : {ANL A,#d}
    Begin DataIn[ACC] := DataIn[ACC] and B2; Inc (PC, 2) End;

    $55 : {ANL A,ad}
    Begin DataIn[ACC] := DataIn[ACC] and DataIn [B2]; Inc (PC, 2) End;

    $56, $57 : {ANL A,@Ri}
    Begin
      If Reg(B1-$56) < $80 then
        DataIn[ACC] := DataIn[ACC] and DataIn [Reg(B1-$56)]
      Else
        DataIn[ACC] := DataIn[ACC] and DataIn2[Reg(B1-$56)];
      Inc (PC)
    End;

    $58..$5F : {ANL A,Rn}
    Begin DataIn[ACC] := DataIn[ACC] and Reg(B1-$58); Inc (PC) End;

    {--------------------------- XRL --------------------------}
    $62 : {XRL ad,A}
    Begin DataIn[B2] := DataIn[B2] xor DataIn[Acc]; Inc (PC, 2) End;

    $63 : {XRL ad,#d}
    Begin DataIn[B2] := DataIn[B2] xor B3; Inc (PC, 3) End;

    $64 : {XRL A,#d}
    Begin DataIn[ACC] := DataIn[ACC] xor B2; Inc (PC, 2) End;

    $65 : {XRL A,ad}
    Begin DataIn[ACC] := DataIn[ACC] xor DataIn [B2]; Inc (PC, 2) End;

    $66, $67 : {XRL A,@Ri}
    Begin
      If Reg(B1-$66) < $80 then
        DataIn[ACC] := DataIn[ACC] xor DataIn [Reg(B1-$66)]
      Else
        DataIn[ACC] := DataIn[ACC] xor DataIn2[Reg(B1-$66)];
      Inc (PC)
    End;

    $68..$6F : {XRL A,Rn}
    Begin DataIn[ACC] := DataIn[ACC] xor Reg(B1-$68); Inc (PC) End;

    {--------------------------- XCH --------------------------}
    $C5 : {XCH A,ad}
    Begin
      I := DataIn [B2];
      DataIn [B2] := DataIn[ACC];
      DataIn[ACC] := I;
      Inc (PC, 2)
    End;

    $C6,$C7 : {XCH A,@Ri}
    Begin
      If Reg(B1-$C6) < $80 then
        I := DataIn [Reg(B1-$C6)]
      Else
        I := DataIn2[Reg(B1-$C6)];
      If Reg(B1-$C6) < $80 then
        DataIn [Reg(B1-$C6)] := DataIn[ACC]
      Else
        DataIn2[Reg(B1-$C6)] := DataIn[ACC];
      DataIn[ACC] := I;
      Inc (PC)
    End;

    $C8..$CF : {XCH A,Rn}
    Begin
      I := Reg(B1-$C8);
      SetReg(B1-$C8, DataIn[ACC]);
      DataIn[ACC] := I;
      Inc (PC)
    End;

    {--------------------------- XCHD -------------------------}
    $D6,$D7 : {XCHD A,@Ri}
    Begin
      If Reg(B1-$D6) < $80 then
        I := DataIn [Reg(B1-$D6)] and $0F
      Else
        I := DataIn2[Reg(B1-$D6)] and $0F;
      If Reg(B1-$D6) < $80 then
        DataIn [Reg(B1-$D6)] :=
          (DataIn [Reg(B1-$D6)] and $F0) or (DataIn[ACC] and $0F)
      Else
        DataIn2[Reg(B1-$D6)] :=
          (DataIn2[Reg(B1-$D6)] and $F0) or (DataIn[ACC] and $0F);
      DataIn[ACC] := (DataIn[ACC] and $F0) or I;
      Inc (PC);
    End;

    {-------------------------- ACALL -------------------------}
    $11,$31,$51,$71,$91,$B1,$D1,$F1 : {ACALL NXXH}
    Begin
      Inc (PC, 2);
      Inc (DataIn[SP]);
      If DataIn[SP] < $80 then
        DataIn [DataIn[SP]] := Lo (PC)
      Else
        DataIn2[DataIn[SP]] := Lo (PC);
      Inc (DataIn[SP]);
      If DataIn[SP] < $80 then
        DataIn [DataIn[SP]] := Hi (PC)
      Else
        DataIn2[DataIn[SP]] := Hi (PC);
      Temp := B1 shr 5; Temp := Swap(Temp) + B2;
      Temp := Temp or (PC and $F800);
      PC := Temp
    End;

    {-------------------------- AJMP --------------------------}
    $01,$21,$41,$61,$81,$A1,$C1,$E1 : {AJMP ?XXXH}
    Begin
      Inc (PC, 2);
      Temp := B1 shr 5; Temp := Swap(Temp) + B2;
      Temp := Temp or (PC and $F800);
      PC := Temp
    End;

    {-------------------------- LJMP --------------------------}
    $02 : {LJMP ad16}
    Begin PC := $100*B2+B3 End;

    {-------------------------- LCALL -------------------------}
    $12 : {LCALL ad16}
    Begin
      Inc (PC, 3);
      Inc (DataIn[SP]);
      If DataIn[SP] < $80 then
        DataIn [DataIn[SP]] := Lo (PC)
      Else
        DataIn2[DataIn[SP]] := Lo (PC);
      Inc (DataIn[SP]);
      If DataIn[SP] < $80 then
        DataIn [DataIn[SP]] := Hi (PC)
      Else
        DataIn2[DataIn[SP]] := Hi (PC);
      PC := $100*B2+B3
    End;

    {--------------------------- RET --------------------------}
    $22 : {RET}
    Begin
      If DataIn[SP] < $80 then
        PC := $100*DataIn [DataIn[SP]]
      Else
        PC := $100*DataIn2[DataIn[SP]];
      Dec (DataIn[SP]);
      If DataIn[SP] < $80 then
        Inc (PC, DataIn [DataIn[SP]])
      Else
        Inc (PC, DataIn2[DataIn[SP]]);
      Dec (DataIn[SP])
    End;

    {--------------------------- RETI -------------------------}
    $32 : {RETI}
    Begin
      If DataIn[SP] < $80 then
        PC := $100*DataIn [DataIn[SP]]
      Else
        PC := $100*DataIn2[DataIn[SP]];
      Dec (DataIn[SP]);
      If DataIn[SP] < $80 then
        Inc (PC, DataIn [DataIn[SP]])
      Else
        Inc (PC, DataIn2[DataIn[SP]]);
      Dec (DataIn[SP])
    End;

    {--------------------------- PUSH -------------------------}
    $C0 : {PUSH ad}
    Begin
      Inc (DataIn[SP]);
      If DataIn[SP] < $80 then
        DataIn [DataIn[SP]] := DataIn [B2]
      Else
        DataIn2[DataIn[SP]] := DataIn [B2];
      Inc (PC, 2)
    End;

    {--------------------------- POP --------------------------}
    $D0 : {POP ad}
    Begin
      If DataIn[SP] < $80 then
        DataIn [B2] := DataIn [DataIn[SP]]
      Else
        DataIn [B2] := DataIn2[DataIn[SP]];
      Dec (DataIn [SP]);
      Inc (PC, 2)
    End;

    {--------------------------- NOP --------------------------}
    $00 : {NOP}
    Begin Inc (PC) End;

    {--------------------------- JBC --------------------------}
    $10 : {JBC bit,rel}
    Begin
      Inc (PC, 3);
      If BitOn (B2) then
        Begin
          ClrBit (B2);
          PC := JR (PC, B3)
        End
    End;

    {---------------------------- JB --------------------------}
    $20 : {JB bit,rel}
    Begin Inc (PC, 3); If BitOn (B2) then PC := JR (PC, B3) End;

    {--------------------------- JNB --------------------------}
    $30 : {JNB bit,rel}
    Begin Inc (PC, 3); If not BitOn (B2) then PC := JR (PC, B3) End;

    {---------------------------- JC --------------------------}
    $40 : {JC rel}
    Begin
      Inc (PC, 2);
      If DataIn[PSW] and CY = CY then PC := JR (PC, B2)
    End;

    {--------------------------- JNC --------------------------}
    $50 : {JNC rel}
    Begin
      Inc (PC, 2);
      If DataIn[PSW] and CY = 0 then PC := JR (PC, B2)
    End;

    {---------------------------- JZ --------------------------}
    $60 : {JZ rel}
    Begin Inc (PC, 2); If DataIn[Acc] = 0 then PC := JR (PC, B2) End;

    {--------------------------- JNZ --------------------------}
    $70 : {JNZ rel}
    Begin Inc (PC, 2); If DataIn[Acc] > 0 then PC := JR (PC, B2) End;

    {--------------------------- SJMP -------------------------}
    $80 : {SJMP rel}
    Begin Inc (PC, 2); PC := JR (PC, B2) End;

    {--------------------------- CJNE -------------------------}
    $B4 : {CJNE A,#d,rel}
    Begin
      Inc (PC, 3);
      If DataIn[Acc] <> B2 then
        Begin
          If DataIn[Acc] < B2 then DataIn[PSW] := DataIn[PSW] or CY
          Else DataIn[PSW] := DataIn[PSW] and ($FF-CY);
          PC := JR (PC, B3)
        End
    End;
    $B5 : {CJNE A,ad,rel}
    Begin
      Inc (PC, 3);
      If DataIn[Acc] <> DataIn [B2] then
        Begin
          If DataIn[Acc] < DataIn [B2] then DataIn[PSW] := DataIn[PSW] or CY
          Else DataIn[PSW] := DataIn[PSW] and ($FF-CY);
          PC := JR (PC, B3)
        End
    End;
    $B6,$B7 : {CJNE @Ri,#d,rel}
    Begin
      Inc (PC, 3);
      If Reg(B1-$B6) < $80 then
        If DataIn [Reg(B1-$B6)] <> B2 then
          Begin
            If DataIn [Reg(B1-$B6)] < B2 then DataIn[PSW] := DataIn[PSW] or CY
            Else DataIn[PSW] := DataIn[PSW] and ($FF-CY);
            PC := JR (PC, B3)
          End
      Else
        If DataIn2[Reg(B1-$B6)] <> B2 then
          Begin
            If DataIn2[Reg(B1-$B6)] < B2 then DataIn[PSW] := DataIn[PSW] or CY
            Else DataIn[PSW] := DataIn[PSW] and ($FF-CY);
            PC := JR (PC, B3)
          End
    End;
    $B8..$BF : {CJNE Rn,#d,rel}
    Begin
      Inc (PC, 3);
      If Reg(B1-$B8) <> B2 then
        Begin
          If Reg(B1-$B8) < B2 then DataIn[PSW] := DataIn[PSW] or CY
          Else DataIn[PSW] := DataIn[PSW] and ($FF-CY);
          PC := JR (PC, B3)
        End
    End;

    {--------------------------- DJNZ -------------------------}
    $D5 : {DJNZ ad,rel}
    Begin
      Dec (DataIn [B2]);
      Inc (PC, 3);
      If DataIn [B2] <> 0 then PC := JR (PC, B3)
    End;
    $D8..$DF : {DJNZ R0,rel}
    Begin
      DecReg(B1-$D8);
      Inc (PC, 2);
      If Reg(B1-$D8) <> 0 then PC := JR (PC, B2)
    End;

    {--------------------------- SETB -------------------------}
    $D2 : {SETB bit}
    Begin SetBit (B2); Inc (PC, 2) End;

    $D3 : {SETB C}
    Begin DataIn[PSW] := DataIn[PSW] or CY; Inc (PC) End;

    {--------------------------- CLR --------------------------}
    $C2 : {CLR bit}
    Begin ClrBit (B2); Inc (PC, 2) End;

    $C3 : {CLR C}
    Begin DataIn[PSW] := DataIn[PSW] and $7F; Inc (PC) End;

    {--------------------------- CPL --------------------------}
    $B2 : {CPL bit}
    Begin CplBit (B2); Inc (PC, 2) End;

    $B3 : {CPL C}
    Begin DataIn[PSW] := DataIn[PSW] xor CY; Inc (PC) End;

    {--------------------------- MOV --------------------------}
    $92 : {MOV bit,C}
    Begin
      If DataIn[PSW] and CY = CY then SetBit (B2)
      Else ClrBit (B2);
      Inc (PC, 2)
    End;

    $A2 : {MOV C,bit}
    Begin
      If BitOn (B2) then DataIn[PSW] := DataIn[PSW] or CY
      Else DataIn[PSW] := DataIn[PSW] and $7F;
      Inc (PC, 2)
    End;

    {--------------------------- ORL --------------------------}
    $A0 : {ORL C,/bit}
    Begin
      Inc (PC, 2);
      If not BitOn (B2) then
        DataIn[PSW] := DataIn[PSW] or CY
    End;
    $72 : {ORL C,bit}
    Begin
      Inc (PC, 2);
      If BitOn (B2) then
        DataIn[PSW] := DataIn[PSW] or CY
    End;

    {--------------------------- ANL --------------------------}
    $B0 : {ANL C,/bit}
    Begin
      Inc (PC, 2);
      If not (not BitOn (B2) and (DataIn[PSW] and CY = CY)) then
        DataIn[PSW] := DataIn[PSW] and ($FF-CY)
    End;
    $82 : {ANL C,bit}
    Begin
      Inc (PC, 2);
      If not (BitOn (B2) and (DataIn[PSW] and CY = CY)) then
        DataIn[PSW] := DataIn[PSW] and ($FF-CY)
    End;

    {--------------------------- JMP --------------------------}
    $73 : {JMP @A+DPTR}
    Begin PC := DataIn[DPL]+256*DataIn[DPH] + DataIn[Acc] End

  End;
  { -------------- ⠭ 䫠 ⭮ -------------- }
  K := DataIn[Acc];
  J := 0;
  For I := 0 to 7 do
    Begin
      K := K shl 1;
      If K and $100 <> 0 then Inc (J)
    End;
  DataIn[PSW] := DataIn[PSW] and ($FF-P);
  If J and 1 <> 0 then DataIn[PSW] := DataIn[PSW] or P;
  { ------------------------------------------------------ }

  { ---------------  稪 横 -------------- }
  Inc (Clk, ClkIncrement[B1]);
  { ------------------------------------------------------ }

  If PC > ProgMax then PC := PC mod (ProgMax+1);
End;

Function Deci (Var H : String) : Word;

  Function S16 (N : Byte) : Word;
  Var
    Z,
    I : Word;
  Begin
    Z := 1;
    If N > 0 then
      For I := 1 to N do
        Z := Z * 16;
    S16 := Z
  End;

Var
  C : Char;
  I : Byte;
  R : Word;
Label
  L;
Begin
  R := 0;
  While Length (H) < 4 do H := '0'+H;
  H := Copy (H, 1, 4);
  For I := 1 to 4 do H[I] := UpCase (H[I]);
  For I := 1 to 4 do
    Begin
      C := H[I];
      If not (C in HexaDec) then Goto L
    End;
  For I := 1 to 4 do
    Begin
      C := H[I];
      If C in Nums then R := R + S16 (4-I)*(Ord(C)-Ord('0'))
      Else R := R + S16 (4-I)*(Ord(C)-Ord('A')+10)
    End;
L:
  Deci := R
End;

Procedure MakeMem;
Var
  I,
  J : Byte;
Begin
  TextColor (Cyan);
  GotoXY (28,16);
  Case MemSel of
    0 : If PageIn = 2 then Write (MsgMemSel[0])
        Else Write (MsgMemSel[1]);
    1 : Write (MsgMemSel[2])
  End;
  TextColor (LightGray);
  For J := 0 to 15 do
    Begin
      GotoXY (33+J*3,16);
      Write  (Hex (J))
    End;
  For J := 0 to 7 do
    Begin
      GotoXY (28,17+J);
      TextColor (LightGray);
      Case MemSel of
        0 : If PageIn <> 2 then
              Write (Hex(Hi(PageIn*128+J*16))+Hex(Lo(PageIn*128+J*16))+' ')
            Else
              Write (Hex(Hi(1*128+J*16))+Hex(Lo(1*128+J*16))+' ');
        1 : Write (Hex(Hi(PageEx*128+J*16))+Hex(Lo(PageEx*128+J*16))+' ')
      End;
      TextColor (Green);
      For I := 0 to 15 do
        Begin
          Case MemSel of
            0 : If PageIn <> 2 then
                  Write (Hex(DataIn[PageIn*128+J*16+I]))
                Else
                  Write (Hex(DataIn2[128+J*16+I]));
            1 : Write (Hex(DataEx^[PageEx*128+J*16+I]))
          End;
          If I <> 15 then Write (' ')
        End
    End
End;

Procedure MakeScr;
Var I : Byte;
Begin
  TextColor (Green);
  GotoXY (2, 16); Write ('R0=' + Hex(Reg(0)));
  GotoXY (2, 17); Write ('R1=' + Hex(Reg(1)));
  GotoXY (2, 18); Write ('R2=' + Hex(Reg(2)));
  GotoXY (2, 19); Write ('R3=' + Hex(Reg(3)));
  GotoXY (2, 20); Write ('R4=' + Hex(Reg(4)));
  GotoXY (2, 21); Write ('R5=' + Hex(Reg(5)));
  GotoXY (2, 22); Write ('R6=' + Hex(Reg(6)));
  GotoXY (2, 23); Write ('R7=' + Hex(Reg(7)));
  TextColor (Cyan );
  GotoXY (2, 24); Write ('RS=',  (DataIn[PSW] and (RS0+RS1) shr 3));
  GotoXY (8, 18); Write (' C=',  (DataIn[PSW] and CY) shr 7);
  GotoXY (8, 19); Write ('AC=',  (DataIn[PSW] and AC) shr 6);
  GotoXY (8, 20); Write ('F0=',  (DataIn[PSW] and F0) shr 5);
  GotoXY (8, 21); Write ('OV=',  (DataIn[PSW] and OV) shr 2);
  GotoXY (8, 22); Write ('F1=',  (DataIn[PSW] and F1) shr 1);
  GotoXY (8, 23); Write (' P=',   DataIn[PSW] and  P);
  GotoXY (7, 24); Write ('PSW='  + Hex(DataIn[PSW]));
  TextColor (LightGray);
  GotoXY (8, 16); Write ('A=' + Hex(DataIn[Acc]));
  GotoXY (8, 17); Write ('B=' + Hex(DataIn[ B ]));
  GotoXY (13,16); Write ('SP=' +  Hex( DataIn[ SP]));
  GotoXY (19,16); Write ('PC=' + Hex(Hi(PC))+Hex(Lo(PC)));
  GotoXY (17,17); Write ('DPTR=' + Hex(DataIn[DPH])+Hex(DataIn[DPL]));
  TextColor (Green);
  GotoXY (13,18); Write ('P0='   + Hex(DataIn[ P0]));
  GotoXY (13,19); Write ('P1='   + Hex(DataIn[ P1]));
  GotoXY (13,20); Write ('P2='   + Hex(DataIn[ P2]));
  GotoXY (13,21); Write ('P3='   + Hex(DataIn[ P3]));
  GotoXY (13,22); Write ('IP='   + Hex(DataIn[IP]));
  GotoXY (13,23); Write ('IE='   + Hex(DataIn[IE]));
  TextColor (Cyan );
  If not RegTog then
    Begin
      GotoXY (19,18); Write ('T0='   + Hex(DataIn[TH0])+Hex(DataIn[TL0]));
      GotoXY (19,19); Write ('T1='   + Hex(DataIn[TH1])+Hex(DataIn[TL1]));
      GotoXY (19,20); Write ('TMOD=' + Hex(DataIn[TMOD]));
      GotoXY (19,21); Write ('TCON=' + Hex(DataIn[TCON]));
      GotoXY (19,22); Write ('SCON=' + Hex(DataIn[SCON]));
      GotoXY (19,23); Write ('SBUF=' + Hex(DataIn[SBUF]));
      GotoXY (19,24); Write ('PCON=' + Hex(DataIn[PCON]))
    End
  Else
    Begin
      GotoXY (19,18); Write ('T2='   + Hex(DataIn[TH2])+Hex(DataIn[TL2]));
      GotoXY (19,19); Write ('RC='   + Hex(DataIn[RCPH])+Hex(DataIn[RCPL]));
      GotoXY (19,20); Write ('T2MD=' + Hex(DataIn[T2MD]));
      GotoXY (19,21); Write ('T2CN=' + Hex(DataIn[T2CN]));
      GotoXY (19,22); Write ('SADR=' + Hex(DataIn[SADR]));
      GotoXY (19,23); Write ('SADN=' + Hex(DataIn[SADN]));
      GotoXY (19,24); Write (' IPH=' + Hex(DataIn[ IPH]))
    End;
  TextColor (Green);
  GotoXY (2,25); Write ('CLK=', Clk:10);
  MakeMem;
  TextColor (DarkGray);
  For I := 2 to 79 do
    Begin
      GotoXY (I, 15); Write ('')
    End
End;

Procedure RegsEdit;
Const
  RIMax : Array [1..36] of Byte =
    (2, 2, 2, 4,
     2, 2, 0, 4,
     2, 1, 2, 4,
     2, 1, 2, 4,
     2, 1, 2, 2,
     2, 1, 2, 2,
     2, 1, 2, 2,
     2, 1, 2, 2,
     1, 2, 0, 2);
  RIX   : Array [1..36] of Byte =
    (5, 10, 16, 22,
     5, 10,  0, 22,
     5, 11, 16, 22,
     5, 11, 16, 22,
     5, 11, 16, 24,
     5, 11, 16, 24,
     5, 11, 16, 24,
     5, 11, 16, 24,
     5, 11,  0, 24);
  RIY   : Array [1..36] of Byte =
    (16, 16, 16, 16,
     17, 17,  0, 17,
     18, 18, 18, 18,
     19, 19, 19, 19,
     20, 20, 20, 20,
     21, 21, 21, 21,
     22, 22, 22, 22,
     23, 23, 23, 23,
     24, 24, 0,  24);
Var
  Code : Integer;
  C : Char;
  S : String;
  K,
  W : Word;
  I, J,
  N : Byte;

  Procedure ReRegs;
  Begin
    Val ('$'+S, W, Code);
    If Code <> 0 then W := 0;
    Case N of
       1  : SetReg(0, W);
       2  : DataIn[Acc]:= W;
       3  : DataIn[SP] := W;
       4  : PC := W;
       5  : SetReg(1, W);
       6  : DataIn[B]  := W;
       8  : Begin DataIn[DPL] := Lo(W); DataIn[DPH] := Hi(W) End;
       9  : SetReg(2, W);
       10 :
         If W = 0 then DataIn[PSW] := DataIn[PSW] and ($FF-CY)
         Else DataIn[PSW] := DataIn[PSW] or CY;
       11 : DataIn[P0] := W;
       12 : If not RegTog then
              Begin DataIn[TL0] := Lo(W); DataIn[TH0] := Hi(W) End
            Else
              Begin DataIn[TL2] := Lo(W); DataIn[TH2] := Hi(W) End;
       13 : SetReg(3, W);
       14 :
         If W = 0 then DataIn[PSW] := DataIn[PSW] and ($FF-AC)
         Else DataIn[PSW] := DataIn[PSW] or AC;
       15 : DataIn[P1] := W;
       16 : If not RegTog then
              Begin DataIn[TL1] := Lo(W); DataIn[TH1] := Hi(W) End
            Else
              Begin DataIn[RCPL] := Lo(W); DataIn[RCPH] := Hi(W) End;
       17 : SetReg(4, W);
       18 :
         If W = 0 then DataIn[PSW] := DataIn[PSW] and ($FF-F0)
         Else DataIn[PSW] := DataIn[PSW] or F0;
       19 : DataIn[P2] := W;
       20 : If not RegTog then
              DataIn[TMOD]:=W
            Else
              DataIn[T2MD]:=W;
       21 : SetReg(5, W);
       22 :
         If W = 0 then DataIn[PSW] := DataIn[PSW] and ($FF-OV)
         Else DataIn[PSW] := DataIn[PSW] or OV;
       23 : DataIn[P3] := W;
       24 : If not RegTog then
              DataIn[TCON]:=W
            Else
              DataIn[T2CN]:=W;
       25 : SetReg(6, W);
       26 :
         If W = 0 then DataIn[PSW] := DataIn[PSW] and ($FF-F1)
         Else DataIn[PSW] := DataIn[PSW] or F1;
       27 : DataIn[IP] := W;
       28 : If not RegTog then
              DataIn[SCON]:=W
            Else
              DataIn[SADR]:=W;
       29 : SetReg(7, W);
       30 :
         If W = 0 then DataIn[PSW] := DataIn[PSW] and ($FF-P)
         Else DataIn[PSW] := DataIn[PSW] or P;
       31 : DataIn[IE] := W;
       32 : If not RegTog then
              DataIn[SBUF]:=W
            Else
              DataIn[SADN]:=W;
       33 :
         If (W = 0) or (W > 3) then
           DataIn[PSW] := DataIn[PSW] and ($FF-RS0-RS1)
         Else DataIn[PSW] := DataIn[PSW] and ($FF-RS0-RS1) or (W shl 3);
       34 : DataIn[PSW]:= W;
       36 : If not RegTog then
              DataIn[PCON]:=W
            Else
              DataIn[ IPH]:=W;
    End;
  End;

Begin
  C := ' ';
  N := RegsPoint;
  CursorOff;
  GotoXY (RIX[N]-1, RIY[N]);
  TextColor (Yellow);
  Write ('');
  While C <> #27 do
    Begin
      If KeyPressed then C := ReadKey;
      If C = #13 then
        Begin
          TextColor (LightRed);
          GotoXY (WhereX-1, WhereY);
          Write ('');
          Write (' ':RIMax[N]);
          GotoXY (WhereX-RIMax[N], WhereY);
          CursorOn (UpL, DnL);
          ReadS (S, RIMax[N]);
          CursorOff;
          C := ' ';
          If N <> 33 then
            Begin
              ReRegs;
              J := 0;
              K := DataIn[Acc];
              For I := 1 to 8 do
                Begin
                  K := K shl 1;
                  If K and $100 = $100 then Inc (J)
                End;
              If J and 1 = 1 then
                DataIn[PSW] := DataIn[PSW] or P
              Else
                DataIn[PSW] := DataIn[PSW] and ($FF-P)
            End
          Else
            Begin
              Val (S, W, Code);
              If (W = 0) or (W > 3) then
                DataIn[PSW] := DataIn[PSW] and ($FF-RS0-RS1)
              Else DataIn[PSW] := DataIn[PSW] and ($FF-RS0-RS1) or (W shl 3);
            End;
          MakeScr;
          TextColor (Yellow);
          GotoXY (RIX[N]-1, RIY[N]);
          Write ('')
        End;
      If C = #0 then
        Begin
          C := ReadKey;
          Case C of
            #72 : {Up}
              Begin
                GotoXY (RIX[N]-1, RIY[N]);
                TextColor (LightGray);
                Write ('=');
                If N > 4 then Dec (N, 4);
                If N = 30 then N := 26;
                If N = 7 then N := 3;
              End;
            #80 : {Down}
              Begin
                GotoXY (RIX[N]-1, RIY[N]);
                TextColor (LightGray);
                Write ('=');
                If N < 33 then Inc (N, 4);
                If N = 30 then N := 34;
                If N =  7 then N := 11;
                If N = 35 then N := 31
              End;
            #75 : {Left}
              Begin
                GotoXY (RIX[N]-1, RIY[N]);
                TextColor (LightGray);
                Write ('=');
                If N mod 4 <> 1 then Dec (N);
                If N =  7 then N :=  6;
                If N = 30 then N := 29;
                If N = 35 then N := 34
              End;
            #77 : {Right}
              Begin
                GotoXY (RIX[N]-1, RIY[N]);
                TextColor (LightGray);
                Write ('=');
                If N mod 4 <> 0 then Inc (N);
                If N = 30 then N := 31;
                If N = 35 then N := 36;
                If N =  7 then N :=  8
              End;
          End;
          GotoXY (RIX[N]-1, RIY[N]);
          TextColor (Yellow);
          Write ('');
        End
    End;
  MakeScr;
  RegsPoint := N
End;

Procedure MemEdit;
Var
  C : Char;
  W : Word;
  N : Byte;
  Code : Integer;
  S : String;
Begin
  N := MemPoint;
  TextColor (Yellow);
  CursorOff;
  GotoXY (32 + (N mod 16)*3, 17 + N div 16);
  Write ('');
  C := ' ';
  While C <> #27 do
    Begin
      If KeyPressed then C := ReadKey;
      If C = #13 then
        Begin
          TextColor (LightRed);
          GotoXY (32 + (N mod 16)*3, 17 + N div 16);
          Write ('  ');
          GotoXY (WhereX-2, WhereY);
          CursorOn (UpL, DnL);
          ReadS (S, 2);
          CursorOff;
          C := ' ';
          Val ('$'+S, W, Code);
          If Code <> 0 then W := 0;
          Case MemSel of
            0 : DataIn [PageIn*128+N] := W;
            1 : DataEx^[PageEx*128+N] := W;
          End;
          MakeScr;
          GotoXY (32 + (N mod 16)*3, 17 + N div 16);
          TextColor (Yellow);
          Write ('')
        End;
      If C = #0 then
        Begin
          C := ReadKey;
          Case C of
            #72 : {Up}
              Begin
                GotoXY (32 + (N mod 16)*3, 17 + N div 16);
                Write (' ');
                If N > 15 then Dec (N, 16);
              End;
            #80 : {Down}
              Begin
                GotoXY (32 + (N mod 16)*3, 17 + N div 16);
                Write (' ');
                If N < $70 then Inc (N, 16);
              End;
            #75 : {Left}
              Begin
                GotoXY (32 + (N mod 16)*3, 17 + N div 16);
                Write (' ');
                If N mod 16 > 0 then Dec (N);
              End;
            #77 : {Right}
              Begin
                GotoXY (32 + (N mod 16)*3, 17 + N div 16);
                Write (' ');
                If N mod 16 < 15 then Inc (N);
              End;
          End;
          GotoXY (32 + (N mod 16)*3, 17 + N div 16);
          TextColor (Yellow);
          Write ('')
        End
    End;
  MakeScr;
  MemPoint := N
End;

Procedure WinAddrEx;
Var
  S : String;
  W : Word;
  Code : Integer;
Begin
  Color (Black, LightRed);
  GotoXY (27, 17);
  Write ('    ');
  CursorOn (UpL, DnL);
  GotoXY (28, 17);
  ReadS (S, 4);
  CursorOff;
  GotoXY (27, 17);
  Write (' ');
  Val ('$'+S, W, Code);
  If (Code <> 0) or (S = '') then
    Begin
      TextColor (LightGray);
      Write (Hex(Hi(PageEx*128))+Hex(Lo(PageEx*128)));
      Exit
    End;
  PageEx := W div 128;
  MakeMem
End;

{-- 頥 . ப -  ⭮⥫쭮  室 --}
Function JRel (PC : Word; R : Byte) : String;
Var
  S : String;
  A : Word;
Begin
  If R < $80 then A := PC+R
  Else A := PC+R-256;
  S := Hex (Hi (A))+Hex (Lo (A));
  JRel := S
End;

{-- 頥 ப -     --}
Function Step (Var PC : Word) : String;
Var
  D  : String;
  B,
  B1,
  B2 : Byte;
  Temp : Word;
Begin
  B  := Prog^[PC];
  B1 := Prog^[PC+1];
  B2 := Prog^[PC+2];

  If B in CAJMP  then D := 'AJMP ' ;
  If B in CACALL then D := 'ACALL ';
  If B in CMOV   then D := 'MOV '  ;
  If B in CINC   then D := 'INC '  ;
  If B in CDEC   then D := 'DEC '  ;
  If B in CADD   then D := 'ADD '  ;
  If B in CADDC  then D := 'ADDC ' ;
  If B in CSUBB  then D := 'SUBB ' ;
  If B in CANL   then D := 'ANL '  ;
  If B in CORL   then D := 'ORL '  ;
  If B in CXRL   then D := 'XRL '  ;
  If B in CCJNE  then D := 'CJNE ' ;
  If B in CDJNZ  then D := 'DJNZ ' ;
  If B in CXCH   then D := 'XCH '  ;
  If B in CXCHD  then D := 'XCHD ' ;
  If B in CCLR   then D := 'CLR '  ;
  If B in CCPL   then D := 'CPL '  ;
  If B in CSETB  then D := 'SETB ' ;
  If B in MOVX   then D := 'MOVX ' ;

  If B =  $00   then D := 'NOP'   ;
  If B =  $C4   then D := 'SWAP A';
  If B =  $D4   then D := 'DA A'  ;
  If B =  $84   then D := 'DIV AB';
  If B =  $A4   then D := 'MUL AB';
  If B =  $22   then D := 'RET'   ;
  If B =  $32   then D := 'RETI'  ;
  If B =  $03   then D := 'RR A'  ;
  If B =  $13   then D := 'RRC A' ;
  If B =  $23   then D := 'RL A'  ;
  If B =  $33   then D := 'RLC A' ;
  If B =  $73   then D := 'JMP @A+DPTR';
  If B =  $83   then D := 'MOVC A,@A+PC';
  If B =  $93   then D := 'MOVC A,@A+DPTR';

  If B =  $E0   then D := D+'A,@DPTR';
  If B =  $E2   then D := D+'A,@R0';
  If B =  $E2   then D := D+'A,@R1';
  If B =  $F0   then D := D+'@DPTR,A';
  If B =  $F2   then D := D+'@R0,A';
  If B =  $F3   then D := D+'@R1,A';

  If B =  $A3   then D := D+'DPTR';

  If B =  $10   then D := 'JBC '  ;
  If B =  $20   then D := 'JB '   ;
  If B =  $30   then D := 'JNB '  ;
  If B =  $40   then D := 'JC '   ;
  If B =  $50   then D := 'JNC '  ;
  If B =  $60   then D := 'JZ '   ;
  If B =  $70   then D := 'JNZ '  ;
  If B =  $80   then D := 'SJMP ' ;
  If B =  $C0   then D := 'PUSH ' ;
  If B =  $D0   then D := 'POP '  ;
  If B =  $02   then D := 'LJMP ' ;
  If B =  $12   then D := 'LCALL ';
  If B =  $A5   then D := '?'     ;

  {------------------------------------}

  D := Hex (Hi (PC))+Hex(Lo (PC))+' '+D;
  If B in OneByte   then Inc (PC)  ;
  If B in TwoByte   then Inc (PC,2);
  If B in ThreeByte then Inc (PC,3);

  {------------------------------------}

  If B in (CAJMP+CACALL) then Begin
    Temp := B shr 5; Temp := Swap(Temp) + B1;
    Temp := Temp or (PC and $F800);
    D := D + Hex(Hi(Temp)) + Hex(Lo(Temp));
  End;

  If B in Accum   then D := D+'A';
  If B in Carry   then D := D+'C';
  If B in Bit_Rel then D := D+Hex(B1)+','+JRel(PC,B2);
  If B in Rel     then D := D+JRel(PC,B1);
  If B in C_Bit   then D := D+'C,'+Hex(B1);
  If B in C_N_Bit then D := D+'C,/'+Hex(B1);
  If B in Bit     then D := D+Hex(B1);
  If B in Ad      then D := D+Hex(B1);
  If B in Ad_A    then D := D+Hex(B1)+',A';
  If B in A_Ad    then D := D+'A,'+Hex(B1);
  If B in Ad16    then D := D+Hex(B1)+Hex(B2);
  If B in Ad_D    then D := D+Hex(B1)+',#'+Hex(B2);
  If B in A_D     then D := D+'A,#'+Hex(B1);
  If B in AdR0    then D := D+'@R0';
  If B in AdR1    then D := D+'@R1';
  If B in A_AdR0  then D := D+'A,@R0';
  If B in A_AdR1  then D := D+'A,@R1';
  If B in A_R0    then D := D+'A,R0';
  If B in A_R1    then D := D+'A,R1';
  If B in A_R2    then D := D+'A,R2';
  If B in A_R3    then D := D+'A,R3';
  If B in A_R4    then D := D+'A,R4';
  If B in A_R5    then D := D+'A,R5';
  If B in A_R6    then D := D+'A,R6';
  If B in A_R7    then D := D+'A,R7';

  If B in RR0     then D := D+'R0';
  If B in RR1     then D := D+'R1';
  If B in RR2     then D := D+'R2';
  If B in RR3     then D := D+'R3';
  If B in RR4     then D := D+'R4';
  If B in RR5     then D := D+'R5';
  If B in RR6     then D := D+'R6';
  If B in RR7     then D := D+'R7';

  If B =  $90     then D := D+'DPTR,#'+Hex(B1)+Hex(B2);
  If B =  $B4     then D := D+'A,#'  +Hex(B1)+','+JRel(PC, B2);
  If B =  $B5     then D := D+'A, '  +Hex(B1)+','+JRel(PC, B2);
  If B =  $B6     then D := D+'@R0,#'+Hex(B1)+','+JRel(PC, B2);
  If B =  $B7     then D := D+'@R1,#'+Hex(B1)+','+JRel(PC, B2);

  If B =  $B8     then D := D+'R0,#' +Hex(B1)+','+JRel(PC, B2);
  If B =  $B9     then D := D+'R1,#' +Hex(B1)+','+JRel(PC, B2);
  If B =  $BA     then D := D+'R2,#' +Hex(B1)+','+JRel(PC, B2);
  If B =  $BB     then D := D+'R3,#' +Hex(B1)+','+JRel(PC, B2);
  If B =  $BC     then D := D+'R4,#' +Hex(B1)+','+JRel(PC, B2);
  If B =  $BD     then D := D+'R5,#' +Hex(B1)+','+JRel(PC, B2);
  If B =  $BE     then D := D+'R6,#' +Hex(B1)+','+JRel(PC, B2);
  If B =  $BF     then D := D+'R7,#' +Hex(B1)+','+JRel(PC, B2);

  If B =  $D5     then D := D+Hex(B1)+','+JRel(PC, B2);

  If B =  $D8     then D := D+'R0,'+JRel(PC, B1);
  If B =  $D9     then D := D+'R1,'+JRel(PC, B1);
  If B =  $DA     then D := D+'R2,'+JRel(PC, B1);
  If B =  $DB     then D := D+'R3,'+JRel(PC, B1);
  If B =  $DC     then D := D+'R4,'+JRel(PC, B1);
  If B =  $DD     then D := D+'R5,'+JRel(PC, B1);
  If B =  $DE     then D := D+'R6,'+JRel(PC, B1);
  If B =  $DF     then D := D+'R7,'+JRel(PC, B1);

  If B =  $F8     then D := D+'R0,A';
  If B =  $F9     then D := D+'R1,A';
  If B =  $FA     then D := D+'R2,A';
  If B =  $FB     then D := D+'R3,A';
  If B =  $FC     then D := D+'R4,A';
  If B =  $FD     then D := D+'R5,A';
  If B =  $FE     then D := D+'R6,A';
  If B =  $FF     then D := D+'R7,A';

  If B =  $A8     then D := D+'R0,'+Hex(B1);
  If B =  $A9     then D := D+'R1,'+Hex(B1);
  If B =  $AA     then D := D+'R2,'+Hex(B1);
  If B =  $AB     then D := D+'R3,'+Hex(B1);
  If B =  $AC     then D := D+'R4,'+Hex(B1);
  If B =  $AD     then D := D+'R5,'+Hex(B1);
  If B =  $AE     then D := D+'R6,'+Hex(B1);
  If B =  $AF     then D := D+'R7,'+Hex(B1);

  If B =  $78     then D := D+'R0,#'+Hex(B1);
  If B =  $79     then D := D+'R1,#'+Hex(B1);
  If B =  $7A     then D := D+'R2,#'+Hex(B1);
  If B =  $7B     then D := D+'R3,#'+Hex(B1);
  If B =  $7C     then D := D+'R4,#'+Hex(B1);
  If B =  $7D     then D := D+'R5,#'+Hex(B1);
  If B =  $7E     then D := D+'R6,#'+Hex(B1);
  If B =  $7F     then D := D+'R7,#'+Hex(B1);

  If B =  $88     then D := D+Hex(B1)+',R0';
  If B =  $89     then D := D+Hex(B1)+',R1';
  If B =  $8A     then D := D+Hex(B1)+',R2';
  If B =  $8B     then D := D+Hex(B1)+',R3';
  If B =  $8C     then D := D+Hex(B1)+',R4';
  If B =  $8D     then D := D+Hex(B1)+',R5';
  If B =  $8E     then D := D+Hex(B1)+',R6';
  If B =  $8F     then D := D+Hex(B1)+',R7';

  If B =  $76     then D := D+'@R0,#'+Hex(B1);
  If B =  $77     then D := D+'@R1,#'+Hex(B1);

  If B =  $85     then D := D+Hex(B2)+','+Hex(B1);

  If B =  $86     then D := D+Hex(B1)+',@R0';
  If B =  $87     then D := D+Hex(B1)+',@R1';

  If B =  $A6     then D := D+'@R0,'+Hex(B1);
  If B =  $A7     then D := D+'@R1,'+Hex(B1);

  If B =  $F6     then D := D+'@R0,A';
  If B =  $F7     then D := D+'@R1,A';

  If B =  $92     then D := D+Hex(B1)+',C';

  Step := D
End;

{ The new keyboard interrupt routine. It will tell us if
  there's a new key pressed on the keyboard. This is
  to make sure ms-windows will not stop running the debugger
  because it tries to get keyboard input from BIOS too often. }

Procedure IRQ1ISR;
Begin
  HasIRQ1Fired := True;
{  POldIRQ1ISR^;}
  inline($9C/$FF/$1E/POldIRQ1ISR); {pushf / call dword ptr [POldIRQ1ISR]}
End;

Procedure Proc;
Const
  Jump : Set of Byte = [
  $02, $10, $20, $30, $40, $50, $60, $70, $80, $73, $22, $32];
Var N : Word;
Begin
  N := PC;
  If (Prog^[PC] in  Jump) or
     (Prog^[PC] in CAJMP) or
     (Prog^[PC] in CCJNE) or
     (Prog^[PC] in CDJNZ) then
       Begin
         OneStep;
         Exit
       End;
  If Prog^[PC] in OneByte then Inc (N, 1);
  If Prog^[PC] in TwoByte then Inc (N, 2);
  If Prog^[PC] in ThreeByte then Inc (N, 3);
  GotoXY (3, 15);
  Color (Black, LightCyan+Blink); Write (MsgEmulating);

  HasIRQ1Fired := False;

  While PC <> N do
    Begin
      OneStep;

      If HasIRQ1Fired then Begin
        HasIRQ1Fired := False;
        If KeyPressed and (ReadKey = ' ') then Exit
      End

    End
End;

Procedure Go;
Var I : Byte;
  Function BreakT (Ad : Word) : Boolean;
  Var
    J : Byte;
    F : Boolean;
  Begin
    F := False;
    For J := 1 to NBreaks do
      If BP [J] = Ad then F := True;
    BreakT := F
  End;
Begin
  If NBreaks = 0 then Exit;
  GotoXY (3, 15);
  Color (Black, LightCyan+Blink); Write (MsgEmulating);

  HasIRQ1Fired := False;

  Repeat
    OneStep;

    If HasIRQ1Fired then Begin
      HasIRQ1Fired := False;
      If KeyPressed and (ReadKey = ' ') then Exit
    End

  Until BreakT (PC)
End;

Begin
  GetCursorLines (UpL, DnL);
  InputCase := 1
End.
